theory Linear_Integer_Inequalities
imports
  Formula
  "$AFP/Show/Show_Instances"
  "$AFP/Derive/Comparator_Generator/Compare_Generator"
  "$AFP/Check_Monad/Check_Monad"
begin


section {* Definition of LIA *}

datatype lin_ineq_type = BoolT | IntegerT
datatype lin_ineq_op = Less | Lesseq | Sum | Const int | Cmult int

derive "show" lin_ineq_op lin_ineq_type

fun lin_ineq_types :: "lin_ineq_op \<Rightarrow> lin_ineq_type list \<times> lin_ineq_type" where
  "lin_ineq_types Less = ([IntegerT,IntegerT], BoolT)"
| "lin_ineq_types Lesseq = ([IntegerT,IntegerT], BoolT)"
| "lin_ineq_types Sum = ([IntegerT,IntegerT], IntegerT)"
| "lin_ineq_types (Const _) = ([], IntegerT)"
| "lin_ineq_types (Cmult _) = ([IntegerT], IntegerT)" (* only admit multiplication with constant *)

type_synonym 'v le_exp = "(lin_ineq_op,'v,lin_ineq_type) exp"
type_synonym 'v le_formula = "(lin_ineq_op,'v,lin_ineq_type) formula"

datatype lin_ineq_value = Boolean (the_bool: bool) | Integer (the_int: int)

fun lin_ineq_type_values :: "lin_ineq_type \<Rightarrow> lin_ineq_value set" where
  "lin_ineq_type_values BoolT = range Boolean"
| "lin_ineq_type_values IntegerT = range Integer"

fun lin_ineq_interpret :: "lin_ineq_op \<Rightarrow> lin_ineq_value list \<Rightarrow> lin_ineq_value" where
  "lin_ineq_interpret Less [x, y] = Boolean (the_int x < the_int y)"
| "lin_ineq_interpret Lesseq [x, y] = Boolean (the_int x \<le> the_int y)"
| "lin_ineq_interpret Sum [x, y] = Integer (the_int x + the_int y)"
| "lin_ineq_interpret (Const x) [] = Integer x"
| "lin_ineq_interpret (Cmult x) [y] = Integer (x * the_int y)"

definition add :: "'v le_exp \<Rightarrow> 'v le_exp \<Rightarrow> 'v le_exp" where "add h t \<equiv> Fun Sum [h,t]"
definition "lia_const c \<equiv> Fun (Const c) []"

definition "LIA = \<lparr>
  types = lin_ineq_types,
  type_values = lin_ineq_type_values,
  bool = BoolT,
  to_bool = the_bool,
  interpret = lin_ineq_interpret
\<rparr>"


subsection {* Flattened version *}

datatype 'v lin_exp = Lin_Exp (lin_const:int) (lin_list:"('v \<times> int)list") (* meaning: c + x1 * c1 + ... xn * cn \<le> 0 *)

fun add_and_mult_list :: "nat \<Rightarrow> 'v lin_exp \<Rightarrow> 'v lin_exp \<Rightarrow> 'v lin_exp" where
"add_and_mult_list c le1 le2 = Lin_Exp (lin_const le1 * (int c) + lin_const le2)
                                 (map (apsnd (op * (int c))) (lin_list le1) @ lin_list le2)"

fun interpret_lin_list :: "('v \<Rightarrow> int) \<Rightarrow> ('v\<times>int) list \<Rightarrow> int"
  where
  "interpret_lin_list f Nil = 0" |
  "interpret_lin_list f (Cons a as) = f (fst a) * (snd a) + interpret_lin_list f as"

fun interpret_lin_exp :: "('v \<Rightarrow> int) \<Rightarrow> 'v lin_exp \<Rightarrow> int"
  where "interpret_lin_exp f le = lin_const le + interpret_lin_list f (lin_list le)"

fun flatten :: "'v le_exp \<Rightarrow> 'v lin_exp" where
"flatten (Var a) = Lin_Exp 0 [(fst a,1)]" |
"flatten (Fun Less [a, b])
  = (case (flatten a,flatten b) of
      (Lin_Exp ac al,Lin_Exp bc bl) \<Rightarrow> Lin_Exp (bc - ac - 1) (map (apsnd (op * (-1))) al @ bl))" |
"flatten (Fun Lesseq [a, b])
  = (case (flatten a,flatten b) of
      (Lin_Exp ac al,Lin_Exp bc bl) \<Rightarrow> Lin_Exp (bc - ac) (map (apsnd (op * (-1))) al @ bl))" |
"flatten (Fun Sum [a, b])
  = (case (flatten a,flatten b) of
      (Lin_Exp ac al,Lin_Exp bc bl) \<Rightarrow> Lin_Exp (bc + ac) (al @ bl))" |
"flatten (Fun (Const a) [])
  = Lin_Exp a []" |
"flatten (Fun (Cmult c) [a])
  = (case (flatten a) of
       (Lin_Exp ac al) \<Rightarrow> Lin_Exp (ac*c) (map (apsnd (op * c)) al))"

fun normalize_sum_duplicates :: "('v \<times> int) list \<Rightarrow> ('v \<times> int) list" where
  "normalize_sum_duplicates [] = []"
| "normalize_sum_duplicates ((x,c) # cxs) = (case partition (\<lambda> (y,d). y = x) cxs of
     (xs, dys) \<Rightarrow> (x,c + listsum (map snd xs)) # normalize_sum_duplicates dys)"

definition normalize_sum_0 :: "('v \<times> int)list \<Rightarrow> ('v \<times> int)list" where
  "normalize_sum_0 = filter (\<lambda> (x,c). c \<noteq> 0)"

definition normalize_sum :: "('v :: linorder \<times> int) list \<Rightarrow> ('v \<times> int) list" where
  "normalize_sum = normalize_sum_0 o sort o normalize_sum_duplicates"

fun compress :: "('v::linorder) lin_exp \<Rightarrow> 'v lin_exp"
where
"compress v = Lin_Exp (lin_const v) (normalize_sum (lin_list v))"

subsection {* Entailment checking functionality *}

fun apply_hint :: "nat list \<Rightarrow> 'v lin_exp list \<Rightarrow> 'v lin_exp" where
  "apply_hint (Cons n ns) (Cons a as) = add_and_mult_list n a (apply_hint ns as)" |
  "apply_hint [] (Cons a as) = add_and_mult_list 1 a (apply_hint [] as)" |
  "apply_hint _ [] = (Lin_Exp 0 [])"

fun negate :: "'v le_exp \<Rightarrow> 'v le_exp"
where
"negate (Fun Less [a,b]) = (Fun Lesseq [b,a])" |
"negate (Fun Lesseq [a,b]) = (Fun Less [b,a])"

fun lin_exp_is_negative_constant :: "'v lin_exp \<Rightarrow> shows check"
where
"lin_exp_is_negative_constant (Lin_Exp a lst)
 =  check (a<0) (shows ''constant remaining must be less than zero'')
 \<then> check (Ball (set lst) (const False)) (shows ''coefficients unequal to zero'')"

fun entailment_checker :: "nat list \<Rightarrow> ('v::linorder) le_exp list \<Rightarrow> 'v le_exp list \<Rightarrow> shows check"
where
"entailment_checker nums_to_multiply preconditions formulas
  = or_ok (lin_exp_is_negative_constant (compress (apply_hint nums_to_multiply (map flatten (preconditions @ map negate formulas)))))
          (lin_exp_is_negative_constant (compress (apply_hint (nums_to_multiply @ [0]) (map flatten (preconditions @ map negate formulas)))))"

locale lin_ineq_theory = "pretheory" LIA
begin

lemma LIA1[simp]:"types = lin_ineq_types" by(simp add:LIA_def)
lemma LIA2[simp]:"type_values = lin_ineq_type_values" by(simp add:LIA_def)
lemma LIA3[simp]:"bool = BoolT" by(simp add:LIA_def)
lemma LIA4[simp]:"to_bool = the_bool" by(simp add:LIA_def)
lemma LIA5[simp]:"interpret = lin_ineq_interpret" by(simp add:LIA_def)

lemma booleanNoInteger [simp]: "\<not> Boolean vb \<in> type_values IntegerT"
by (simp add: image_iff)
lemma integerNoBoolean [simp]: "\<not> Integer vb \<in> type_values BoolT"
by (simp add: image_iff)

lemma interpret_append[simp]:
  "interpret_lin_list f (a @ b) = interpret_lin_list f a + interpret_lin_list f b"
by(induct a,simp_all)

lemma interpret_mult[simp]:
  "interpret_lin_list f (map (apsnd (op * c)) x) = c * interpret_lin_list f x"
by(induct x,auto simp:distrib_left)

lemma add_and_mult_list_ok[simp]:
  assumes a1:"0 \<le> interpret_lin_exp f l1" and a2:"0 \<le> interpret_lin_exp f l2"
  shows "0 \<le> interpret_lin_exp f (add_and_mult_list n l1 l2)"
proof -
  have "0 \<le> int n * (lin_const l1 + interpret_lin_list f (lin_list l1))"
    using a1 by simp
  then show ?thesis using a2 by (simp add: distrib_right mult_of_nat_commute)
qed

declare add_and_mult_list.simps[simp del]

sublocale "theory" LIA
proof
  show "typelist_match ds (fst (types f)) \<Longrightarrow> interpret f ds \<in> type_values (snd (types f))" for f ds
  by(cases f; induct f ds rule: lin_ineq_interpret.induct, auto)
  show "type_values ty \<noteq> {}" for ty by(cases ty, auto)
qed

lemma lin_ineq_theory_induct[consumes 2, case_names Less Lesseq Sum Const Cmult Var]:
  fixes P :: "(lin_ineq_op, 'a \<times> lin_ineq_type) Term.term \<Rightarrow> lin_ineq_value \<Rightarrow> bool"
  assumes assignmentOK:"assignment \<alpha>"
  and welltyped:"type_of a \<in> range Some"
  and "\<And> ta tb ia ib.
        eval \<alpha> ta = Integer ia \<Longrightarrow> eval \<alpha> tb = Integer ib 
       \<Longrightarrow> type_of ta = Some IntegerT \<Longrightarrow> type_of tb = Some IntegerT
       \<Longrightarrow> P ta (Integer ia) \<Longrightarrow> P tb (Integer ib)
       \<Longrightarrow> P (Fun Less [ta,tb]) (Boolean (ia < ib))" 
  and "\<And> ta tb ia ib.
        eval \<alpha> ta = Integer ia \<Longrightarrow> eval \<alpha> tb = Integer ib 
       \<Longrightarrow> type_of ta = Some IntegerT \<Longrightarrow> type_of tb = Some IntegerT
       \<Longrightarrow> P ta (Integer ia) \<Longrightarrow> P tb (Integer ib)
       \<Longrightarrow> P (Fun Lesseq [ta,tb]) (Boolean (ia \<le> ib))" 
  and "\<And> ta tb ia ib.
        eval \<alpha> ta = Integer ia \<Longrightarrow> eval \<alpha> tb = Integer ib 
       \<Longrightarrow> type_of ta = Some IntegerT \<Longrightarrow> type_of tb = Some IntegerT
       \<Longrightarrow> P ta (Integer ia) \<Longrightarrow> P tb (Integer ib)
       \<Longrightarrow> P (Fun Sum [ta,tb]) (Integer (ia + ib))" 
  and "\<And> x. P (Fun (Const x) []) (Integer x)"
  and "\<And> ta ia x.
        eval \<alpha> ta = Integer ia
       \<Longrightarrow> type_of ta = Some IntegerT
       \<Longrightarrow> P ta (Integer ia)
       \<Longrightarrow> P (Fun (Cmult x) [ta]) (Integer (x * ia))"
  and "\<And> x y. P (Var (x,y)) (\<alpha> (x,y))"
  shows "P a (eval \<alpha> a)"
using assms(2-) proof(induct a)
  case (Var x) obtain a b where "(a,b) = x" using prod.collapse by blast
    thus ?case using Var.prems by auto next
  case (Fun f ls)
  have types_ts: "mapM type_of ls = Some (fst (lin_ineq_types f))"
    using type_of_fun(2) notin_range_Some Fun.prems(1) unfolding LIA1 by blast
  show ?case using assignmentOK proof(cases f)
    case Less then obtain ta tb where ab[simp]:"ls = [ta,tb]" 
      and at:"type_of ta = Some IntegerT" and bt:"type_of tb = Some IntegerT"
      using mapM_Some[OF types_ts] by auto
      obtain ai bi where ai[simp]: "eval \<alpha> ta = Integer ai" and bi[simp]:"eval \<alpha> tb = Integer bi"
        using eval_types[OF at assms(1)] eval_types[OF bt assms(1)] by auto
      have pa:"P ta (Integer ai)" using Fun.hyps[of ta] ab at Fun.prems by simp
      have pb:"P tb (Integer bi)" using Fun.hyps[of tb] ab bt Fun.prems by simp
      show ?thesis using Fun.prems(2)[OF ai bi at bt pa pb] by (simp add: Less) next
    case Lesseq then obtain ta tb where ab[simp]:"ls = [ta,tb]" 
      and at:"type_of ta = Some IntegerT" and bt:"type_of tb = Some IntegerT"
      using mapM_Some[OF types_ts] by auto
      obtain ai bi where ai[simp]: "eval \<alpha> ta = Integer ai" and bi[simp]:"eval \<alpha> tb = Integer bi"
        using eval_types[OF at assms(1)] eval_types[OF bt assms(1)] by auto
      have pa:"P ta (Integer ai)" using Fun.hyps[of ta] ab at Fun.prems by simp
      have pb:"P tb (Integer bi)" using Fun.hyps[of tb] ab bt Fun.prems by simp
      show ?thesis using Fun.prems(3)[OF ai bi at bt pa pb] by (simp add: Lesseq) next
    case Sum then obtain ta tb where ab[simp]:"ls = [ta,tb]" 
      and at:"type_of ta = Some IntegerT" and bt:"type_of tb = Some IntegerT"
      using mapM_Some[OF types_ts] by auto
      obtain ai bi where ai[simp]: "eval \<alpha> ta = Integer ai" and bi[simp]:"eval \<alpha> tb = Integer bi"
        using eval_types[OF at assms(1)] eval_types[OF bt assms(1)] by auto
      have pa:"P ta (Integer ai)" using Fun.hyps[of ta] ab at Fun.prems by simp
      have pb:"P tb (Integer bi)" using Fun.hyps[of tb] ab bt Fun.prems by simp
      show ?thesis using Fun.prems(4)[OF ai bi at bt pa pb] by (simp add: Sum) next
    case (Const x)
      hence la[simp]: "ls = []" using mapM_Some[OF types_ts] by auto
      show ?thesis using Fun.prems(5) by (simp add: Const) next
    case (Cmult x) then obtain ta where ab[simp]:"ls = [ta]" 
      and at:"type_of ta = Some IntegerT"
      using mapM_Some[OF types_ts] by auto
      obtain ai where ai[simp]: "eval \<alpha> ta = Integer ai"
        using eval_types[OF at assms(1)] by auto
      have pa:"P ta (Integer ai)" using Fun.hyps[of ta] ab at Fun.prems by simp
      show ?thesis using Fun.prems(6)[OF ai at pa] by (simp add: Cmult)
  qed
qed

lemma flatten_agrees_int:
  assumes "assignment \<alpha>"
  and "type_of a = Some IntegerT"
  shows "eval \<alpha> a = Integer (interpret_lin_exp (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (flatten a))"
proof -
  have "assignment \<alpha> \<Longrightarrow> type_of a \<in> range Some \<Longrightarrow> type_of a = Some IntegerT \<Longrightarrow> ?thesis"
    proof(induct \<alpha> a rule:lin_ineq_theory_induct)
      case (Var x t) thus ?case
      by (simp,metis assms(1)[unfolded assignment_def] booleanNoInteger lin_ineq_value.exhaust_sel)
    qed(simp_all add:lin_exp.case_eq_if distrib_left)
  thus ?thesis using assms by simp
qed

lemma flatten_agrees_bool:
  assumes "assignment \<alpha>"
  and "is_bool \<phi>"
  shows "the_bool (eval \<alpha> \<phi>) = (interpret_lin_exp (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (flatten \<phi>) \<ge> 0)"
proof -
  have "assignment \<alpha> \<Longrightarrow> type_of \<phi> \<in> range Some \<Longrightarrow> is_bool \<phi> \<Longrightarrow> ?thesis"
    proof(induct \<alpha> \<phi> rule:lin_ineq_theory_induct)
      case (Lesseq ta tb ia ib) show ?case
        using Lesseq.hyps(1)[unfolded flatten_agrees_int[OF assms(1) Lesseq.hyps(3)]]
              Lesseq.hyps(2)[unfolded flatten_agrees_int[OF assms(1) Lesseq.hyps(4)]]
        by(auto simp:lin_exp.case_eq_if) next
      case (Less ta tb ia ib) show ?case
        using Less.hyps(1)[unfolded flatten_agrees_int[OF assms(1) Less.hyps(3)]]
              Less.hyps(2)[unfolded flatten_agrees_int[OF assms(1) Less.hyps(4)]]
        by(auto simp:lin_exp.case_eq_if)
    qed(simp_all add:is_bool_def)
  from this[OF assms(1)] show ?thesis using assms unfolding is_bool_def by simp
qed


lemma partition_listsum: "partition f xs = (yes,no) \<Longrightarrow> listsum (map g xs) = 
  listsum (map g yes) + (listsum (map g no) :: 'a :: ab_group_add)"
proof (induct xs arbitrary: yes no)
  case (Cons x xs)
  obtain y n where p: "partition f xs = (y,n)" by force
  from Cons(2)[unfolded partition.simps p Let_def split]
  have id: "(yes, no) = (if f x then (x # y, n) else (y, x # n))" by auto
  from Cons(1)[OF p] have IH: "listsum (map g xs) = listsum (map g y) + listsum (map g n)" .
  show ?case using id by (cases "f x", auto simp: IH ac_simps)
qed simp

lemma normalize_sum_duplicates[simp]: 
  "(\<Sum>(x, c)\<leftarrow> normalize_sum_duplicates cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
proof (induct cxs rule: normalize_sum_duplicates.induct)
  case (2 x c cxs)
  obtain xs dys where part: "partition (\<lambda> (y,d). y = x) cxs = (xs,dys)" by force
  hence xs: "\<And> y c. (y,c) \<in> set xs \<Longrightarrow> y = x" by auto
  show ?case unfolding normalize_sum_duplicates.simps part split
    by (simp add: field_simps partition_listsum[OF part] 2[OF part[symmetric]],
    insert xs, induct xs, (force simp: field_simps)+)
qed simp

lemma normalize_sum_0: "(\<Sum>(x, c)\<leftarrow> normalize_sum_0 cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
  unfolding normalize_sum_0_def by (induct cxs, auto)

lemma normalize_sum_duplicates_distinct: 
  "distinct (map fst (normalize_sum_duplicates cxs)) \<and> fst ` set (normalize_sum_duplicates cxs) = fst ` set cxs"
proof (induct cxs rule: normalize_sum_duplicates.induct)
  case (2 x c cxs)
  obtain xs dys where part: "partition (\<lambda> (y,d). y = x) cxs = (xs,dys)" by force
  hence dys: "\<And> y c. (y,c) \<in> set dys \<Longrightarrow> y \<noteq> x" by auto
  show ?case using 2[OF part[symmetric]] unfolding normalize_sum_duplicates.simps part split
    using part by auto
qed simp

lemma normalize_sum: 
  "(\<Sum>(x, c)\<leftarrow> normalize_sum cxs. c * \<alpha> x) = (\<Sum>(x, c) \<leftarrow> cxs. c * \<alpha> x)"
proof -
  from normalize_sum_duplicates_distinct[of cxs]
  have dist: "distinct (normalize_sum_duplicates cxs)"
    by (simp add: distinct_map)
  show ?thesis
    unfolding normalize_sum_def o_def normalize_sum_0
    by (subst setsum.distinct_set_conv_list[symmetric], simp add: dist, simp,
    subst setsum.distinct_set_conv_list[OF dist], simp)
qed

lemma interpret_lin_list_connection: "(\<Sum>(x, c)\<leftarrow> cxs. c * \<alpha> x) = interpret_lin_list \<alpha> cxs"
by(induct cxs,auto)

lemma normalize_ok: "interpret_lin_list \<alpha> (normalize_sum cxs) = interpret_lin_list \<alpha> cxs"
using normalize_sum by (auto simp:interpret_lin_list_connection)

lemma compress_ok: "interpret_lin_exp \<alpha> (compress cxs) = interpret_lin_exp \<alpha> cxs"
using normalize_ok by auto

(* definitions to avoid the need for existential quantification *)
fun hd2 where "hd2 [] = 1" | "hd2 (Cons a _) = a"
fun tl2 where "tl2 [] = []" | "tl2 (Cons _ a) = a"

lemma apply_hint_adds:
  "apply_hint h (Cons a as) = add_and_mult_list (hd2 h) a (apply_hint (tl2 h) as)"
by(cases h,auto)

lemma apply_hint_ok:
  assumes "Ball (set lst) (\<lambda> le. 0 \<le> interpret_lin_exp f le)"
  shows "0 \<le> interpret_lin_exp f (apply_hint h lst)"
using assms proof(induct lst arbitrary:h)
  case (Cons a as)
  hence as_ok: "\<And> h. 0 \<le> interpret_lin_exp f (apply_hint h as)" by auto
  have a_ok: "0 \<le> interpret_lin_exp f a" using Cons.prems by auto
  show ?case unfolding apply_hint_adds using add_and_mult_list_ok[OF a_ok as_ok] by blast
qed(simp)

lemma negate_negates :
  assumes "assignment \<alpha>"
  and "is_bool \<phi>"
  shows "eval_true \<alpha> (negate \<phi>) \<longleftrightarrow> \<not> eval_true \<alpha> \<phi>"
proof -
  have "assignment \<alpha> \<Longrightarrow> type_of \<phi> \<in> range Some \<Longrightarrow> is_bool \<phi> \<Longrightarrow> ?thesis"
    by(induct \<alpha> \<phi> rule:lin_ineq_theory_induct,auto simp:eval_true_def is_bool_def)
  thus ?thesis using assms unfolding is_bool_def by simp
qed

(* Show that there exists an assignment \<alpha> *)
fun some_alpha::"'a \<times> lin_ineq_type \<Rightarrow> lin_ineq_value" where
  "some_alpha (v,IntegerT) = Integer 0" |
  "some_alpha (v,BoolT) = Boolean True"
lemma some_alpha:"assignment some_alpha" proof -
  {fix x t have "some_alpha (x::'a, t) \<in> lin_ineq_type_values t" by(cases t;simp)}
  thus ?thesis by(auto simp:assignment_def)
  qed

lemma negate_preserves_is_bool[simp]:
  assumes "is_bool (\<phi> :: (lin_ineq_op, 'a \<times> lin_ineq_type) Term.term)"
  shows "is_bool (negate \<phi>)"
proof -
  {fix \<alpha>::"'a \<times> lin_ineq_type \<Rightarrow> lin_ineq_value"
  have "assignment \<alpha> \<Longrightarrow> type_of \<phi> \<in> range Some \<Longrightarrow> is_bool \<phi> \<Longrightarrow> ?thesis"
    by (induct \<alpha> \<phi> rule:lin_ineq_theory_induct,auto simp:is_bool_def)}
  thus ?thesis using assms some_alpha unfolding is_bool_def by simp
qed

lemma lin_exp_is_negative_ok :
  assumes "isOK (lin_exp_is_negative_constant (cxs :: 'a lin_exp))"
  and "assignment (\<alpha> :: 'a \<times> lin_ineq_type \<Rightarrow> lin_ineq_value)"
  shows "\<not> (0 \<le> interpret_lin_exp (\<lambda>a. the_int (\<alpha> (a, IntegerT))) cxs)"
proof(cases cxs) case (Lin_Exp c r)
  have c0:"c < 0" using assms unfolding Lin_Exp by auto
  have "Ball (set r) (const False)" using assms unfolding Lin_Exp by auto
  hence r:"r = []" by fastforce
  have "0 > (interpret_lin_exp (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (Lin_Exp c r))" by(simp add: r c0)
  thus ?thesis by (simp add: Lin_Exp)
qed

lemma entailment_checker_content_checks_for_False:
  assumes "isOK (lin_exp_is_negative_constant (compress (apply_hint h (map flatten \<phi>s))))"
  and "assignment \<alpha>"
  and "\<forall> \<phi> \<in> set \<phi>s. is_bool \<phi> \<and> eval_true \<alpha> \<phi>"
  shows "False"
proof -
  have ths:
    "\<not> (0 \<le> interpret_lin_exp (\<lambda>a. the_int (\<alpha> (a, IntegerT))) (apply_hint h (map flatten \<phi>s)))"
    using lin_exp_is_negative_ok[OF assms(1,2)] compress_ok[of "\<lambda>a. the_int (\<alpha> (a, IntegerT))"]
    by simp
  have allHold:
    "Ball (set (map flatten \<phi>s)) (\<lambda> le. 0 \<le> interpret_lin_exp (\<lambda>a. the_int (\<alpha> (a, IntegerT))) le)"
    using assms(3) flatten_agrees_bool[OF assms(2)] unfolding eval_true_def by auto
  show ?thesis using ths apply_hint_ok[OF allHold] by auto
qed


lemma entailment_checker_ok:
  assumes "isOK (entailment_checker h prems concls)"
  " set prems \<subseteq> BOOL_EXP"
  " set concls \<subseteq> BOOL_EXP"
  " assignment \<alpha>"
  " \<forall>prem \<in> set prems. eval_true \<alpha> prem"
  shows " \<exists>concl \<in> set concls. eval_true \<alpha> concl"
proof -
  let ?inner = "prems @ map negate concls"
  obtain h2 where isok:"isOK (lin_exp_is_negative_constant
              (compress (apply_hint h2 (map flatten ?inner))))"
    using assms(1) or_is_or unfolding entailment_checker.simps by blast
  have "\<forall> v \<in> set ?inner. eval_true \<alpha> v \<Longrightarrow> False"
    using entailment_checker_content_checks_for_False[OF isok assms(4)] assms(2,3)
          negate_preserves_is_bool by auto
  hence "Ball (set (map negate concls)) (\<lambda> v. eval_true \<alpha> v) \<Longrightarrow> False" using assms(5) by auto
  hence "Ball (set concls) (\<lambda> v. \<not> eval_true \<alpha> v) \<Longrightarrow> False"
    using negate_negates[OF assms(4)] assms(3) by auto
  thus ?thesis by auto
qed

lemma theory_checker: "theory_checker LIA entailment_checker"
 by(insert lin_ineq_theory.entailment_checker_ok,unfold_locales)
end
declare "lin_ineq_theory.LIA1"[simp]
declare "lin_ineq_theory.LIA2"[simp]
declare "lin_ineq_theory.LIA3"[simp]
declare "lin_ineq_theory.LIA4"[simp]
declare "lin_ineq_theory.LIA5"[simp]

end
