(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2013, 2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2012-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Non_Inf_Order
imports
  "../Rewriting/Trs"
  "$AFP/Abstract-Rewriting/SN_Order_Carrier"
begin

text \<open>We define non infinitesmal orders and related notions
  as in the CADE07 "Bounded Increase" paper.\<close>

text \<open>for monotonicity, we do not use 0, 1, -1, and 2 as in CADE07, but a dedicated datatype\<close>

datatype dependance = Ignore | Increase | Decrease | Wild

type_synonym 'f dep = "('f \<times> nat) \<Rightarrow> nat \<Rightarrow> dependance"

fun rel_dep :: "'a rel \<Rightarrow> dependance \<Rightarrow> 'a rel" (infix "^^^" 52) where
  "_ ^^^ Ignore = {}"
| "r ^^^ Increase = r"
| "r ^^^ Decrease = (r ^-1)"
| "r ^^^ Wild = (r^<->)"

lemma rel_dep_mono: "r1 \<subseteq> r2 \<Longrightarrow> r1 ^^^ k \<subseteq> r2 ^^^ k"
  by (cases k, auto)

fun invert_dep :: "dependance \<Rightarrow> dependance" where
  "invert_dep Increase = Decrease"
| "invert_dep Decrease = Increase"  
| "invert_dep d = d"
  
lemma rel_dep_invert_mono: "r1^-1 \<subseteq> r2 \<Longrightarrow> r1 ^^^ invert_dep k \<subseteq> r2 ^^^ k"
  by (cases k, auto)

text \<open>for compatibility (monotonicity) we only consider F-terms, not all terms as in CADE07\<close>

definition dep_compat :: "'f sig \<Rightarrow> ('f, 'v) term rel \<Rightarrow> 'f dep \<Rightarrow> bool" where
  "dep_compat F r \<pi> \<longleftrightarrow> (\<forall> f bef s t aft. \<Union>(funas_term ` ({s,t} \<union> set bef \<union> set aft)) \<subseteq> F \<longrightarrow> 
  {(s,t)} ^^^ \<pi> (f,Suc (length bef + length aft)) (length bef) \<subseteq> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r)"

lemma dep_compatE[elim]: assumes dep: "dep_compat F r \<pi>" shows "funas_args_term (Fun f (bef @ s # aft)) \<subseteq> F \<Longrightarrow> funas_term t \<subseteq> F \<Longrightarrow> 
  {(s,t)} ^^^ \<pi> (f,Suc (length bef + length aft)) (length bef) \<subseteq> r
  \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r"
  unfolding funas_args_term_def
  by (rule dep[unfolded dep_compat_def, rule_format, of s t bef aft f], force+)

lemma dep_compatI[intro]: assumes "\<And> f bef s t aft. 
  \<lbrakk>\<And> u. u \<in> {s,t} \<union> set bef \<union> set aft \<Longrightarrow> funas_term u \<subseteq> F\<rbrakk> \<Longrightarrow>
  {(s,t)} ^^^ \<pi> (f,Suc (length bef + length aft)) (length bef) \<subseteq> r
  \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r"
  shows "dep_compat F r \<pi>" 
  unfolding dep_compat_def
  by (intro allI impI, rule assms, auto)

text \<open>F-stability (closure under F-substitutions) is the same as in CADE07\<close>  

definition F_subst_closed :: "'f sig \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool" where
  "F_subst_closed F r \<longleftrightarrow>
    (\<forall> \<sigma> s t. \<Union>(funas_term ` range \<sigma>) \<subseteq> F \<longrightarrow> (s,t) \<in> r \<longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> r)"

lemma F_subst_closedI[intro]:
  assumes "\<And> \<sigma> s t. \<Union>(funas_term ` range \<sigma>) \<subseteq> F \<Longrightarrow> (s,t) \<in> r \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> r"
  shows "F_subst_closed F r" using assms unfolding F_subst_closed_def by blast

lemma F_subst_closedD[elim]: "F_subst_closed F r \<Longrightarrow> \<Union>(funas_term ` range \<sigma>) \<subseteq> F \<Longrightarrow> (s,t) \<in> r \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> r"
  unfolding F_subst_closed_def by auto

lemma F_subst_closed_UNIV:
  "F_subst_closed F r \<Longrightarrow> F = UNIV \<Longrightarrow> subst.closed r" unfolding F_subst_closed_def
  using subst.closedI[of r] by auto

text \<open>a locale to have all nice properties together\<close>

locale non_inf_order = fixes S NS :: "'a rel"
  assumes trans_NS: "trans NS"
  and refl_NS: "refl NS"
  and compat: "S O NS \<subseteq> S"
  and compat2: "NS O S \<subseteq> S"
  and trans_S: "trans S"
  and non_inf: "non_inf S"
begin

lemma compat_rtrancl: assumes ab: "(a,b) \<in> S"
  and bc: "(b,c) \<in> (NS \<union> S)^*"
  shows "(a,c) \<in> S"
  using bc
proof (induct)
  case base
  show ?case by (rule ab)
next
  case (step c d)
  from step(2-3) show ?case using compat trans_S unfolding trans_def by blast
qed

lemma chain_split: fixes f :: "'b seq" and m :: "'b \<Rightarrow> 'a" 
  assumes fS: "fS \<subseteq> {i. (m (f i), m (f (Suc i))) \<in> S}"
  and fB: "fB \<subseteq> {i. (m (f i), a) \<in> NS \<union> S}" 
  and decr: "\<And> i. (m (f i), m (f (Suc i))) \<in> NS \<union> S" 
  shows "(\<forall>\<^sub>\<infinity> j. j \<notin> fS) \<or> (\<forall>\<^sub>\<infinity> j. j \<notin> fB)"
proof (cases "(\<exists>\<^sub>\<infinity> i. i \<in> fS) \<and> (\<exists>\<^sub>\<infinity> i. i \<in> fB)")
  case False
  thus ?thesis unfolding INFM_nat_le MOST_nat_le by blast
next
  case True
  def g \<equiv> "\<lambda> i. m (f i)"
  from decr have orient': "\<And> i. (g i, g (Suc i)) \<in> NS \<union> S" unfolding g_def by auto
  have Ps': "\<And> i. i \<in> fS \<Longrightarrow> (g i, g (Suc i)) \<in> S" using fS unfolding g_def by auto
  have Pb': "\<And> i. i \<in> fB \<Longrightarrow> (g i, a) \<in> NS \<union> S" using fB unfolding g_def by auto
  from decr have orient': "\<And> i. (g i, g (Suc i)) \<in> NS \<union> S" unfolding g_def by auto
  {
    fix i j :: nat
    assume "i \<le> j"
    hence "j = (j - i) + i" by arith
    then obtain k where j: "j = k + i" by blast
    have "(g i, g j) \<in> (NS \<union> S)^*" unfolding j
    proof (induct k)
      case (Suc k)
      with orient'[of "k + i"] have "(g i, g (Suc k + i)) \<in> (NS \<union> S)^* O (NS \<union> S)" by auto
      thus ?case by regexp
    qed simp
  } note NS = this

  let ?P = "\<lambda> i. (g i, g (Suc i)) \<in> S"
  (* since for non-inf. we require strict decrease and boundedness in every step, 
     we need to compress our sequences, i.e., we consider a subsequence where all 
     non-strict decreases are dropped *)
  from True have infm: "INFM i. ?P i" using fS unfolding g_def INFM_nat_le by blast
  interpret infinitely_many ?P
    by (unfold_locales, rule infm)
  let ?i = "\<lambda> i. index i"
  def v \<equiv> "\<lambda> i. g (?i i)"
  {
    fix i
    from index_p[of i] have S: "(v i, g (Suc (?i i))) \<in> S" unfolding v_def .
    from index_ordered[of i] have "Suc (?i i) \<le> ?i (Suc i)" by simp
    from compat_rtrancl[OF S NS[OF this]] have vv: "(v i, v (Suc i)) \<in> S" unfolding v_def .
    from True obtain j where ji: "j \<ge> ?i (Suc i)" and Pb: "j \<in> fB" unfolding INFM_nat_le by auto
    from fB Pb have *: "(g j, a) \<in> NS \<union> S" unfolding g_def by auto
    from compat_rtrancl[OF vv[unfolded v_def] NS[OF ji]] have "(v i, g j) \<in> S" unfolding v_def .
    with * compat trans_S[unfolded trans_def] have "(v i, a) \<in> S" by auto
    hence "(v i, v (Suc i)) \<in> S" and "(v i, a) \<in> S" using vv by blast+
  } 
  (* using this compressed sequence, we have direct contradiction to non-inf. *)
  with non_infE[OF non_inf, of v a] have False by blast
  thus ?thesis ..
qed

end

locale non_inf_order_trs = non_inf_order S NS
  for S :: "('f,'v)trs"
  and NS :: "('f,'v)trs" +
  fixes F :: "'f sig"
  and \<pi> :: "'f dep"
  assumes stable_NS: "F_subst_closed F NS"
  and stable_S: "F_subst_closed F S"
  and dep_compat_NS: "dep_compat F NS \<pi>"

end

