(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2013-2016)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2016)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Sets of Unifiers\<close>

theory Unifiers
imports 
  Term
begin

type_synonym ('f, 'v) equation = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) equations = "('f, 'v) equation set"

text \<open>The set of unifiers for a given set of equations.\<close>
definition unifiers :: "('f, 'v) equations \<Rightarrow> ('f, 'v) subst set" where
  "unifiers E = {\<sigma>. \<forall>p\<in>E. fst p \<cdot> \<sigma> = snd p \<cdot> \<sigma>}"

text \<open>Applying a substitution to a set of equations.\<close>
definition subst_set :: "('f, 'v) subst \<Rightarrow> ('f, 'v) equations \<Rightarrow> ('f, 'v) equations"
where
  "subst_set \<sigma> E = (\<lambda>e. (fst e \<cdot> \<sigma>, snd e \<cdot> \<sigma>)) ` E"

text \<open>Check whether a substitution is a most-general unifier (mgu) of a set of equations.\<close>
definition is_mgu :: "('f, 'v) subst \<Rightarrow> ('f, 'v) equations \<Rightarrow> bool" where
  "is_mgu \<sigma> E \<longleftrightarrow> \<sigma> \<in> unifiers E \<and> (\<forall>\<tau> \<in> unifiers E. (\<exists>\<gamma>. \<tau> = \<sigma> \<circ>\<^sub>s \<gamma>))"

lemma is_mguI:
  fixes \<sigma> :: "('f, 'v) subst"
  assumes "\<forall>(s, t) \<in> E. s \<cdot> \<sigma> = t \<cdot> \<sigma>"
    and "\<And>\<tau> :: ('f, 'v) subst. \<forall>(s, t) \<in> E. s \<cdot> \<tau> = t \<cdot> \<tau> \<Longrightarrow> \<exists>\<gamma> :: ('f, 'v) subst. \<tau> = \<sigma> \<circ>\<^sub>s \<gamma>"
  shows "is_mgu \<sigma> E"
  using assms by (fastforce simp: is_mgu_def unifiers_def)

text \<open>Check whether a set of equations is unifiable.\<close>
definition "unifiable E \<longleftrightarrow> (\<exists>\<sigma>. \<sigma> \<in> unifiers E)"

text \<open>The following property characterizes idempotent mgus, i.e., mgus
@{term \<sigma>} for which @{term "\<sigma> \<circ>\<^sub>s \<sigma> = \<sigma>"}.\<close>
definition is_imgu :: "('f, 'v) subst \<Rightarrow> ('f, 'v) equations \<Rightarrow> bool" where
  "is_imgu \<sigma> E \<longleftrightarrow> \<sigma> \<in> unifiers E \<and> (\<forall>\<tau> \<in> unifiers E. \<tau> = \<sigma> \<circ>\<^sub>s \<tau>)"

lemma is_imgu_imp_is_mgu:
  assumes "is_imgu \<sigma> E"
  shows "is_mgu \<sigma> E"
  using assms by (auto simp: is_imgu_def is_mgu_def)

lemma member_unifiersE [elim]:
  "\<lbrakk>\<sigma> \<in> unifiers E; (\<And>e. e \<in> E \<Longrightarrow> fst e \<cdot> \<sigma> = snd e \<cdot> \<sigma>) \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
  by (force simp: unifiers_def)

lemma unifiers_set_zip [simp]:
  assumes "length ss = length ts"
  shows "unifiers (set (zip ss ts)) = {\<sigma>. map (\<lambda>t. t \<cdot> \<sigma>) ss = map (\<lambda>t. t \<cdot> \<sigma>) ts}"
  using assms by (induct ss ts rule: list_induct2) (auto simp: unifiers_def)

lemma unifiers_Fun [simp]:
  "\<sigma> \<in> unifiers {(Fun f ss, Fun g ts)} \<longleftrightarrow>
    length ss = length ts \<and> f = g \<and> \<sigma> \<in> unifiers (set (zip ss ts))"
  by (auto simp: unifiers_def dest: map_eq_imp_length_eq)
     (induct ss ts rule: list_induct2, simp_all)

lemma unifiers_Un [simp]:
  "unifiers (s \<union> t) = unifiers s \<inter> unifiers t"
  by (auto simp: unifiers_def)

lemma unifiers_insert: -- "simp not added for readability (and termination)"
  "unifiers (insert p t) = {\<sigma>. fst p \<cdot> \<sigma> = snd p \<cdot> \<sigma>} \<inter> unifiers t"
  by (auto simp: unifiers_def)

lemma unifiers_insert_swap:
  "unifiers (insert (s, t) E) = unifiers (insert (t, s) E)"
  by (auto simp: unifiers_insert)

lemma unifiers_insert_Var_swap [simp]:
  "unifiers (insert (t, Var x) E) = unifiers (insert (Var x, t) E)"
  by (rule unifiers_insert_swap)

lemma unifiers_empty [simp]:
  "unifiers {} = UNIV"
  by (auto simp: unifiers_def)

lemma unifiers_insert_ident [simp]:
  "unifiers (insert (t, t) E) = unifiers E" 
  by (auto simp: unifiers_insert)

lemma subst_set_insert [simp]:
  "subst_set \<sigma> (insert e E) = insert (fst e \<cdot> \<sigma>, snd e \<cdot> \<sigma>) (subst_set \<sigma> E)"
  by (auto simp: subst_set_def)

lemma unifiers_subst_set [simp]:
  "\<tau> \<in> unifiers (subst_set \<sigma> E) \<longleftrightarrow> \<sigma> \<circ>\<^sub>s \<tau> \<in> unifiers E"
  by (auto simp: unifiers_def subst_set_def)

lemma unifiers_insert_Var_left:
  "\<sigma> \<in> unifiers (insert (Var x, t) E) \<Longrightarrow> \<sigma> \<in> unifiers (subst_set (subst x t) E)"
  by (auto simp: unifiers_def subst_set_def)

lemma unifiers_insert_VarD:
  shows "\<sigma> \<in> unifiers (insert (Var x, t) E) \<Longrightarrow> subst x t \<circ>\<^sub>s \<sigma> = \<sigma>"
    and "\<sigma> \<in> unifiers (insert (t, Var x) E) \<Longrightarrow> subst x t \<circ>\<^sub>s \<sigma> = \<sigma>"
  using assms by (auto simp: unifiers_def)

lemma is_mgu_empty [simp]:
  "is_mgu Var {}"
  by (auto simp: is_mgu_def)

lemma is_mgu_insert_trivial [simp]:
  "is_mgu \<sigma> (insert (t, t) E) = is_mgu \<sigma> E"
  by (auto simp: is_mgu_def)

lemma is_mgu_insert_decomp [simp]:
  assumes "length ss = length ts"
  shows "is_mgu \<sigma> (insert (Fun f ss, Fun f ts) E) \<longleftrightarrow>
    is_mgu \<sigma> (E \<union> set (zip ss ts))"
  using assms by (auto simp: is_mgu_def unifiers_insert)

lemma is_mgu_subst_set_subst:
  assumes "x \<notin> vars_term t"
    and "is_mgu \<sigma> (subst_set (subst x t) E)" (is "is_mgu \<sigma> ?E")
  shows "is_mgu (subst x t \<circ>\<^sub>s \<sigma>) (insert (Var x, t) E)" (is "is_mgu ?\<sigma> ?E'")
proof -
  from \<open>is_mgu \<sigma> ?E\<close>
    have "?\<sigma> \<in> unifiers E"
    and *: "\<forall>\<tau>. (subst x t \<circ>\<^sub>s \<tau>) \<in> unifiers E \<longrightarrow> (\<exists>\<mu>. \<tau> = \<sigma> \<circ>\<^sub>s \<mu>)"
    by (auto simp: is_mgu_def)
  then have "?\<sigma> \<in> unifiers ?E'" using assms by (simp add: unifiers_insert subst_compose)
  moreover have "\<forall>\<tau>. \<tau> \<in> unifiers ?E' \<longrightarrow> (\<exists>\<mu>. \<tau> = ?\<sigma> \<circ>\<^sub>s \<mu>)"
  proof (intro allI impI)
    fix \<tau>
    assume **: "\<tau> \<in> unifiers ?E'"
    then have [simp]: "subst x t \<circ>\<^sub>s \<tau> = \<tau>" by (blast dest: unifiers_insert_VarD)
    from unifiers_insert_Var_left [OF **]
      have "subst x t \<circ>\<^sub>s \<tau> \<in> unifiers E" by (simp)
    with * obtain \<mu> where "\<tau> = \<sigma> \<circ>\<^sub>s \<mu>" by blast
    then have "subst x t \<circ>\<^sub>s \<tau> = subst x t \<circ>\<^sub>s \<sigma> \<circ>\<^sub>s \<mu>" by (auto simp: ac_simps)
    then show "\<exists>\<mu>. \<tau> = subst x t \<circ>\<^sub>s \<sigma> \<circ>\<^sub>s \<mu>" by auto
  qed
  ultimately show "is_mgu ?\<sigma> ?E'" by (simp add: is_mgu_def)
qed

lemma is_mgu_insert_swap:
  "is_mgu \<sigma> (insert (s, t) E) = is_mgu \<sigma> (insert (t, s) E)"
  by (auto simp: is_mgu_def unifiers_def)

lemma is_mgu_insert_Var_swap [simp]:
  "is_mgu \<sigma> (insert (t, Var x) E) = is_mgu \<sigma> (insert (Var x, t) E)"
  by (rule is_mgu_insert_swap)
 
lemma unifiable_UnD [dest]:
  "unifiable (M \<union> N) \<Longrightarrow> unifiable M \<and> unifiable N"
  by (auto simp: unifiable_def)

lemma supt_imp_not_unifiable:
  assumes "s \<rhd> t"
  shows "\<not> unifiable {(t, s)}"
proof
  assume "unifiable {(t, s)}"
  then obtain \<sigma> where "\<sigma> \<in> unifiers {(t, s)}"
    by (auto simp: unifiable_def)
  then have "t \<cdot> \<sigma> = s \<cdot> \<sigma>" by (auto)
  moreover have "s \<cdot> \<sigma> \<rhd> t \<cdot> \<sigma>"
    using assms by (metis instance_no_supt_imp_no_supt)
  ultimately show False by auto
qed

lemma unifiable_insert_swap:
  "unifiable (insert (s, t) E) = unifiable (insert (t, s) E)"
  by (auto simp: unifiable_def unifiers_insert_swap)

lemma unifiable_insert_Var_swap [simp]:
  "unifiable (insert (t, Var x) E) \<longleftrightarrow> unifiable (insert (Var x, t) E)"
  by (rule unifiable_insert_swap)

lemma unifiers_occur_left_is_Fun:
  fixes t :: "('f, 'v) term"
  assumes "x \<in> vars_term t" and "is_Fun t"
  shows "unifiers (insert (Var x, t) E) = {}"
proof (rule ccontr)
  have "t \<rhd> Var x" using assms by auto
  from supt_subst [OF this]
    have "\<forall>\<sigma>::('f, 'v) subst. t \<cdot> \<sigma> \<noteq> Var x \<cdot> \<sigma>" by (auto simp: supt_supteq_conv)
  moreover assume "\<not> ?thesis"
  ultimately show False by (simp add: unifiers_def) metis
qed

lemma unifiers_occur_left_not_Var:
  "x \<in> vars_term t \<Longrightarrow> t \<noteq> Var x \<Longrightarrow> unifiers (insert (Var x, t) E) = {}"
  using unifiers_occur_left_is_Fun [of x t] by (cases t) simp_all

lemma unifiers_occur_left_Fun:
  "x \<in> (\<Union>t\<in>set ts. vars_term t) \<Longrightarrow> unifiers (insert (Var x, Fun f ts) E) = {}"
    using unifiers_occur_left_is_Fun [of x "Fun f ts"] by simp

lemmas unifiers_occur_left_simps [simp] =
  unifiers_occur_left_is_Fun
  unifiers_occur_left_not_Var
  unifiers_occur_left_Fun

lemma unifiers_Int1 [simp]:
  "(s, t) \<in> E \<Longrightarrow> unifiers {(s, t)} \<inter> unifiers E = unifiers E"
  by (auto simp: unifiers_def)

lemma imgu_linear_var_disjoint:
  assumes "is_imgu \<sigma> {(l2 |_ p, l1)}"
  and "p \<in> poss l2"
  and "linear_term l2"
  and "vars_term l1 \<inter> vars_term l2 = {}"
  and "q \<in> poss l2"
  and "parallel_pos p q"
  shows "l2 |_ q = l2 |_ q \<cdot> \<sigma>"
using assms
proof (induct p arbitrary: q l2)
  case (PCons i p)
  from this(3) obtain f ls where 
    l2[simp]: "l2 = Fun f ls" and 
    i: "i < length ls" and 
    p: "p \<in> poss (ls ! i)"
      by (cases l2) (auto)
  then have l2i: "l2 |_ i <# p = ls ! i |_ p" by auto
  have "linear_term (ls ! i)" using PCons(4) l2 i by simp
  moreover have "vars_term l1 \<inter> vars_term (ls ! i) = {}" using PCons(5) l2 i by force
  ultimately have IH: "\<And>q. q \<in> poss (ls ! i) \<Longrightarrow> p \<bottom> q \<Longrightarrow> ls ! i |_ q = ls ! i |_ q \<cdot> \<sigma>" 
    using PCons(1)[OF PCons(2)[unfolded l2i] p] by blast
  from PCons(7) obtain j q' where q: "q = j <# q'" by (cases q) auto
  show ?case
  proof (cases "j = i") 
    case True with PCons(6,7) IH q show ?thesis by simp
  next
    case False
    from PCons(6) q have j: "j < length ls" by simp 
    { fix y
      assume y: "y \<in> vars_term (l2 |_ q)"
      let ?\<tau> = "\<lambda>x. if x = y then Var y else \<sigma> x"
      from y PCons(6) q j have yj:"y \<in> vars_term (ls ! j)" 
        by simp (meson subt_at_imp_supteq subteq_Var_imp_in_vars_term supteq_Var supteq_trans)
      { fix i j
        assume j:"j < length ls" and i:"i < length ls" and neq: "i \<noteq> j"
        from j PCons(4) have "\<forall>i < j. vars_term (ls ! i) \<inter> vars_term (ls ! j) = {}"
          by (auto simp : is_partition_def)
        moreover from i PCons(4) have "\<forall>j < i. vars_term (ls ! i) \<inter> vars_term (ls ! j) = {}"
          by (auto simp : is_partition_def)
        ultimately have "vars_term (ls ! i) \<inter> vars_term (ls ! j) = {}" 
          using neq by (cases "i < j") auto
      }
      from this[OF i j False] have "y \<notin> vars_term (ls ! i)" using yj by auto
      then have "y \<notin> vars_term (l2 |_ i <# p)"
        by (metis l2i p subt_at_imp_supteq subteq_Var_imp_in_vars_term supteq_Var supteq_trans)
      hence "\<forall>x \<in> vars_term (l2 |_ i <# p). ?\<tau> x = \<sigma> x" by auto
      hence l2\<tau>\<sigma>: "l2 |_ i <# p \<cdot> ?\<tau> = l2 |_ i <# p \<cdot> \<sigma>" using term_subst_eq[of _ \<sigma> ?\<tau>] by simp
      from PCons(5) have "y \<notin> vars_term l1" using y PCons(6) vars_term_subt_at by fastforce
      then have "\<forall>x \<in> vars_term l1. ?\<tau> x = \<sigma> x" by auto
      then have l1\<tau>\<sigma>:"l1 \<cdot> ?\<tau> = l1 \<cdot> \<sigma>" using term_subst_eq[of _ \<sigma> ?\<tau>] by simp
      have "l1 \<cdot> \<sigma> = l2 |_ i <# p \<cdot> \<sigma>" using PCons(2) unfolding is_imgu_def by auto
      then have "l1 \<cdot> ?\<tau> = l2 |_ i <# p \<cdot> ?\<tau>"  using l1\<tau>\<sigma> l2\<tau>\<sigma> by simp
      then have "?\<tau> \<in> unifiers {(l2 |_ i <# p, l1)}" unfolding unifiers_def by simp
      with PCons(2) have \<tau>\<sigma>:"?\<tau> = \<sigma> \<circ>\<^sub>s ?\<tau>" unfolding is_imgu_def by blast
      have "Var y = Var y \<cdot> \<sigma>"
      proof (rule ccontr)
        let ?x = "Var y \<cdot> \<sigma>"
        assume *:"Var y \<noteq> ?x"
        have "Var y = Var y \<cdot> ?\<tau>" by auto
        also have "... = (Var y \<cdot> \<sigma>) \<cdot> ?\<tau>" using \<tau>\<sigma> subst_subst by metis 
        finally have xy:"?x \<cdot> \<sigma> = Var y" using * by (cases "\<sigma> y") auto 
        have "\<sigma> \<circ>\<^sub>s \<sigma> = \<sigma>" using PCons(2) unfolding is_imgu_def by auto
        then have "?x \<cdot> (\<sigma> \<circ>\<^sub>s \<sigma>) = Var y" using xy by auto
        moreover have "?x \<cdot> \<sigma> \<cdot> \<sigma> = ?x" using xy by auto
        ultimately show False using * by auto
      qed
      }
    then show ?thesis by (simp add: term_subst_eq)
  qed
qed auto

end

